/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package org.w3c.rdf.model.impl;

import org.w3c.rdf.model.*;
import org.w3c.rdf.util.RDFDigestUtil;
import org.w3c.tools.crypt.*;

import java.util.*;

/**
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */

public class RDFModelImpl implements RDFModel, Digestable {

  //  static LongTable resources = new LongTable(); // long -> ResourceImpl
  //  static LongTable literals = new LongTable(); // long -> LiteralImpl

  //  static Hashtable resources = new Hashtable();
  //  static Hashtable literals = new Hashtable();

  // whether triples and lookup are shared
  boolean shared = false;
  Digest digest;

  //  int findCounter = 0;

  String uri;
  Hashtable triples;
  //  LongTable subjects, predicates, objects;

  // first find operation creates lookup table
  LongTable lookup;

  public RDFModelImpl() {
    triples = new Hashtable();
    lookup = new LongTable(true);
  }

  protected RDFModelImpl(String uri, Hashtable triples, LongTable lookup, boolean shared) {
    this.uri = uri;
    this.triples = triples;
    this.lookup = lookup;
    this.shared = shared;
  }

  /**
   * Set a base URI for the message.
   * Affects creating of new resources and serialization syntax.
   * <code>null</code> is just alright, meaning that
   * the serializer will use always rdf:about and never rdf:ID
   */
  public void setSourceURI(String uri) {
    this.uri = uri;
  }

  /**
   * Returns current base URI setting.
   */
  public String getSourceURI() {
    return uri;
  }

  // RDFModel access

  /**
   * Number of triples in the model
   *
   * @return  number of triples
   */
  public int size() {
    return triples.size();
  }

  /**
   * Enumerate triples
   */
  public Enumeration elements() {
    return triples.elements();
  }

  /**
   * Tests if the model contains the given triple.
   *
   * @return  <code>true</code> if the triple belongs to the model;
   *          <code>false</code> otherwise.
   * @see     org.w3c.rdf.Triple
   */
  public synchronized boolean contains(Triple t) {
    return triples.contains(t);
  }

  // Model manipulation: add, remove, find

  /**
   * Adds a new triple to the model. A literal is created out of the string parameter <code>object</code>.
   * This method is just a shortcut.
   *
   * @see     org.w3c.rdf.Resource
   */
  public synchronized void add(Resource subject, Resource predicate, String object) {
    add(createTriple(subject, predicate, createLiteral(object)));
  }

  /**
   * Adds a new triple to the model. The object of the triple is an RDFNode.
   *
   * @see     org.w3c.rdf.Resource
   * @see     org.w3c.rdf.RDFNode
   */
  public synchronized void add(Resource subject, Resource predicate, RDFNode object) {

    add(createTriple(subject, predicate, object));
  }

  long getLookupValue(Resource subject, Resource predicate, RDFNode object) {

    int s1 = subject != null ? subject.hashCode() : 0;
    int s2 = predicate != null ? predicate.hashCode() : 0;
    int s3 = object != null ? object.hashCode() : 0;
    long l = ((long)s1 * 37 + s2) * 37 + s3;
    //    long l = (((((long)s1) << 16) ^ s2) << 16) ^ s3;
    //    System.err.println("lookup is " + l + " for " + subject + "," + predicate + "," + object);
    return l;
  }

  void addLookup(Triple t) {

    Resource s = t.subject();
    Resource p = t.predicate();
    RDFNode o = t.object();

    lookup.put(getLookupValue(null, null, o),    t);
    lookup.put(getLookupValue(null, p,    null), t);
    lookup.put(getLookupValue(null, p,    o),    t);
    lookup.put(getLookupValue(s,    null, null), t);
    lookup.put(getLookupValue(s,    null, o),    t);
    lookup.put(getLookupValue(s,    p,    null), t);
    lookup.put(getLookupValue(s,    p,    o),    t);
  }

  void removeLookup(Triple t) {

    Resource s = t.subject();
    Resource p = t.predicate();
    RDFNode o = t.object();

    lookup.remove(getLookupValue(null, null, o),    t);
    lookup.remove(getLookupValue(null, p,    null), t);
    lookup.remove(getLookupValue(null, p,    o),    t);
    lookup.remove(getLookupValue(s,    null, null), t);
    lookup.remove(getLookupValue(s,    null, o),    t);
    lookup.remove(getLookupValue(s,    p,    null), t);
    lookup.remove(getLookupValue(s,    p,    o),    t);
  }

  /**
   * Adds a new triple to the model.
   *
   * @see     org.w3c.rdf.Triple
   */
  public synchronized void add( Triple t ) {

    /*
    subjects.put(t.subject().hashCode(), t);
    predicates.put(t.predicate().hashCode(), t);
    objectes.put(t.object().hashCode(), t);
    */

    makeNonShared();

    triples.put(t, t);

    if(validLookup())
      addLookup(t);
  }

  void makeNonShared() {

    if(shared) {
      triples = (Hashtable)triples.clone();
      lookup = (LongTable)lookup.clone();
      shared = false;
    }
  }

  boolean validLookup() {

    return lookup.size() > 0;
  }

  /**
   * Removes the triple from the model.
   *
   * @see     org.w3c.rdf.Triple
   */
  public synchronized void remove(Triple t) {

    makeNonShared();

    /*
    TripleImpl.DEBUG = true;
    System.err.println("Removing triple " + t + " from the table, before: " + triples.size());
    for(Enumeration en = triples.keys(); en.hasMoreElements();) {
      Triple tt = (Triple)en.nextElement();
      System.err.println("TRIPLE " + tt.hashCode() + " against " + t.hashCode() + ": " + tt.equals(t));
    }
    */
    triples.remove(t);
    //    System.err.println("After " + triples.size());

    if(validLookup())
      removeLookup(t);
  }


  /**
   * General method to search for triples.
   * <code>null</code> input for any parameter will match anything.
   * <p>Example: <code>Model result = m.find( null, RDF.type, new Resource("http://...#MyClass") )</code>
   * <p>finds all instances of the class <code>MyClass</code>
   *
   * @see     org.w3c.rdf.Resource
   * @see     org.w3c.rdf.RDFNode
   */
  public synchronized RDFModel find( Resource subject, Resource predicate, RDFNode object ) {

    // TODO: make everything synchronized by triples

    //    findCounter++;

    RDFModel res = new RDFModelImpl();

    if(triples.size() == 0)
      return res;

    if(subject == null && predicate == null && object == null)
      return this.duplicate();

    /*
    int max = triples.size();
    int s1 = subject != null ? subjects.multigetsize(subject.hashCode()) : max;
    int s2 = predicates != null ? predicates.multigetsize(predicate.hashCode()) : max;
    int s3 = object != null ? objects.multigetsize(object.hashCode()) : max;
    */

    //    long l = getLookupValue(subject, predicate, object);
    //    System.err.println("LOOKING FOR " + subject+","+predicate+","+object + "=" + l + ": " + lookup.multigetsize(l));

    //    int counter = 0;

    boolean createLookup = !validLookup();

    //    if(createLookup)
    //      System.err.println("Creating lookup, size=" + triples.size() + ", shared=" + shared);

    Enumeration en = createLookup ? triples.elements() : lookup.multiget(getLookupValue(subject, predicate, object));

    while(en.hasMoreElements()) {

      Triple t = (Triple)en.nextElement();
      //      System.err.println("CHECKING: " + t);

      if(createLookup)
	addLookup(t);

      if(matchTriple(t, subject, predicate, object))
	res.add(t);
      //      counter++;
    }
    //    System.err.println("TRIPLES CHECKED: " + counter);
    return res;
  }

  /**
   * Search for a single triple.
   * <code>null</code> input for any parameter will match anything.
   *
   * @return  a single triple if such was found;<br>
   *          <code>null</code> if nothing was found;<br>
   *          otherwise a RuntimeException is thrown.
  public synchronized Triple find1( Resource subject, Resource predicate, RDFNode object ) {

    return (Triple)find(subject, predicate, object).elements().nextElement();
   */

    /*
    if(subject == null && predicate == null && object == null)
      return triples.size() == 0 ? null : (Triple)triples.elements().nextElement();

    Model res = new RDFModelImpl();
    for(Enumeration en = lookup.multiget(getLookupValue(subject, predicate, object)); en.hasMoreElements();) {

      Triple t = (Triple)en.nextElement();
      if(matchTriple(t, subject, predicate, object))
	return t;
    }
    return null;
    */
  //  }

  boolean matchTriple(Triple triple, Resource subject, Resource predicate, RDFNode object) {

    if( subject != null && !triple.subject().equals(subject))
      return false;
      
    if(predicate != null && !triple.predicate().equals(predicate))
      return false;
      
    if(object != null && !triple.object().equals(object))
      return false;

    return true;
  }

  /**
   * Clone the model.
   */
  public RDFModel duplicate() {

    //    return new RDFModelImpl(uri, (Hashtable)triples.clone(), lookup != null ? (LongTable)lookup.clone() : null, true);

    // creates a model that shares triples and lookup with this model
    return new RDFModelImpl(uri, triples, lookup, true);
  }

  public Object clone() {
    return duplicate();
  }

  /**
   * Creates empty model of the same Class
   */
  public RDFModel create() {
    return new RDFModelImpl();
  }

  /**
   * Creates a resource out of a string
   */
  public Resource createResource(String str) {

    return new ResourceImpl(str);

    /*
    Resource r = (Resource)resources.get(str);
    if(r == null)
      resources.put(str, r = new ResourceImpl(str));
    //    System.err.println("--- REUSE: RESOURCE");
    return r;
    */

    //    return new SlimResource(Crc64.compute(str), str);
    /*
    long hash = Crc64.compute(str);
    Resource r = (Resource)resources.get(hash);
    if(r == null)
      resources.put(hash, r = new SlimResource(hash, str));
    return r;
    */
  }

  /**
   * Creates a literal out of a string
   */
  public Literal createLiteral(String str) {

    return new LiteralImpl(str);

    /*
    Literal l = (Literal)literals.get(str);
    if(l == null)
      literals.put(str, l = new LiteralImpl(str, isXML));
    //    System.err.println("--- REUSE: LITERAL");
    return l;
    */

    //    return new SlimLiteral(Crc64.compute(str), str);
    //    return new LiteralImpl(str, isXML);
    /*
    long hash = Crc64.compute(str);
    Literal l = (Literal)literals.get(hash);
    if(l == null)
      literals.put(hash, l = new SlimLiteral(hash, str));
    return l;
    */
  }

  /*
  Resource getResource(Resource r) {
    if(r instanceof SlimResource)
      return r;
    else
      return createResource(r.toString());
  }

  Literal getLiteral(Literal l) {
    if(l instanceof SlimLiteral)
      return l;
    else
      return createLiteral(l.toString(), false); // FIXME
  }
  */

  /**
   * Creates a triple
   */
  public Triple createTriple(Resource subject, Resource predicate, RDFNode object) {

    return new TripleImpl(predicate, subject, object);
  }


  // Resource implementation

  public String getLabel() {

    return getURI();
  }

  public Digest getDigest() {

    if(digest == null) {
      digest = RDFDigestUtil.getModelDigest(this);
    }
    return digest;
  }

  public String getURI() {

    return "urn:rdf:model:" + DigestUtil.toHexString(getDigest());
  }

  /*
  protected void finalize() throws Throwable {
    System.err.println("Model " + this + " finalized: triples " + triples.size() + ", find ops: " + findCounter);
  }
  */

}
